# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

# The JSON to object mapping was originally created with assistance from OpenAI's ChatGPT.
# For more information about ChatGPT, visit https://openai.com/

from dataclasses import dataclass, field
from typing import Optional


@dataclass
class ProtocolDiscoveryDJson:
    ec: Optional[int] = field(default=None)
    bu: Optional[str] = field(default=None)
    pf: Optional[str] = field(default=None)
    dip: Optional[int] = field(default=None)
    errorMsg: Optional[str] = None


@dataclass
class IotcDiscoveryResponseJson:
    d: ProtocolDiscoveryDJson = field(default_factory=ProtocolDiscoveryDJson)
    status: Optional[int] = field(default=None)
    message: Optional[str] = field(default=None)



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

import json
import urllib.parse
from typing import Final, Union

from avnet.iotconnect.sdk.lite import DeviceConfig, DeviceConfigError
from avnet.iotconnect.sdk.sdklib.protocol.discovery import IotcDiscoveryResponseJson
from avnet.iotconnect.sdk.sdklib.protocol.identity import ProtocolIdentityPJson, ProtocolMetaJson, ProtocolIdentityResponseJson
from avnet.iotconnect.sdk.sdklib.util import deserialize_dataclass


class DeviceIdentityData:
    def __init__(self, mqtt: ProtocolIdentityPJson, metadata: ProtocolMetaJson):
        self.host = mqtt.h
        self.client_id = mqtt.id
        self.username = mqtt.un
        self.topics = mqtt.topics

        self.pf = metadata.pf
        self.is_edge_device = metadata.edge
        self.is_gateway_device = metadata.gtw
        self.protocol_version = str(metadata.v)


class DraDiscoveryUrl:
    method: str = "GET"  # To clarify that get should be used to parse the response
    API_URL_FORMAT: Final[str] = "%s/api/v2.1/dsdk/cpId/%s/env/%s"

    def get_api_url(self) -> str:
        return DraDiscoveryUrl.API_URL_FORMAT % (
            self.config.discovery_url,
            urllib.parse.quote(self.config.cpid, safe=''),
            urllib.parse.quote(self.config.env, safe='')
        )

    def __init__(self, config: DeviceConfig):
        self.config = config


class DraIdentityUrl:
    UID_API_URL_FORMAT: Final[str] = "%s/uid/%s"

    def __init__(self, base_url):
        self.base_url = base_url

    method: str = "GET"  # To clarify that get should be used to parse the response

    def get_uid_api_url(self, config: DeviceConfig) -> str:
        return DraIdentityUrl.UID_API_URL_FORMAT % (
            self.base_url,
            urllib.parse.quote(config.duid, safe='')
        )

    def _validate_identity_response(self, ird: ProtocolIdentityResponseJson):
        # TODO: validate and throw DeviceConfigError
        pass


class DraDeviceInfoParser:
    EC_RESPONSE_MAPPING = [
        "OK – No Error",
        "Device not found. Device is not whitelisted to platform.",
        "Device is not active.",
        "Un-Associated. Device has not any template associated with it.",
        "Device is not acquired. Device is created but it is in release state.",
        "Device is disabled. It’s disabled from broker by Platform Admin",
        "Company not found as SID is not valid",
        "Subscription is expired.",
        "Connection Not Allowed.",
        "Invalid Bootstrap Certificate.",
        "Invalid Operational Certificate."
    ]

    @classmethod
    def _parsing_common(cls, what: str, rd: Union[IotcDiscoveryResponseJson, ProtocolIdentityResponseJson]):
        """ Helper to parse either discovery or identity response common error fields """

        ec_message = 'not available'
        has_error = False
        if rd.d is not None:
            if rd.d.ec != 0:
                has_error = True
                if rd.d.ec <= len(cls.EC_RESPONSE_MAPPING):
                    ec_message = 'ec=%d (%s)' % (rd.d.ec, cls.EC_RESPONSE_MAPPING[rd.d.ec])
                else:
                    ec_message = 'ec==%d' % rd.d.ec
        else:
            has_error = True

        if rd.status != 200:
            has_error = True

        if has_error:
            raise DeviceConfigError(
                '%s failed. Error: "%s" status=%d message=%s' % (
                    what,
                    ec_message,
                    rd.status,
                    rd.message or "(message not available)"
                )
            )

    @classmethod
    def parse_discovery_response(cls, discovery_response: str) -> str:
        """ Parses discovery response JSON and Returns base URL or raises DeviceConfigError """

        drd: IotcDiscoveryResponseJson
        try:
            drd = deserialize_dataclass(IotcDiscoveryResponseJson, json.loads(discovery_response))
        except json.JSONDecodeError as json_error:
            raise DeviceConfigError("Discovery JSON Parsing Error: %s" % str(json_error))
        cls._parsing_common("Discovery", drd)

        if drd.d.bu is None:
            raise DeviceConfigError("Discovery response is missing base URL")

        return drd.d.bu

    @classmethod
    def parse_identity_response(cls, identity_response: str) -> DeviceIdentityData:
        ird: ProtocolIdentityResponseJson
        try:
            ird = deserialize_dataclass(ProtocolIdentityResponseJson, json.loads(identity_response))
        except json.JSONDecodeError as json_error:
            raise DeviceConfigError("Identity JSON Parsing Error: %s" % str(json_error))
        cls._parsing_common("Identity", ird)

        return DeviceIdentityData(ird.d.p, ird.d.meta)



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

# The JSON to object mapping was originally created with assistance from OpenAI's ChatGPT.
# For more information about ChatGPT, visit https://openai.com/

# This file contains inbound message types

from dataclasses import dataclass, field
from typing import Optional, List

@dataclass
class ProtocolC2dMessageJson:
    ct: Optional[int] = field(default=None)
    ack: Optional[str] = field(default=None)
    df: Optional[int] = field(default=None)
    f: Optional[int] = field(default=None)

@dataclass
class ProtocolCommandMessageJson:
    ct: Optional[int] = field(default=None)
    cmd: Optional[str] = field(default=None)
    ack: Optional[str] = field(default=None)


@dataclass
class ProtocolOtaUrlJson:
    url: str = field(default=None)
    fileName: str = field(default=None)


@dataclass
class ProtocolOtaMessageJson:
    ct: Optional[int] = field(default=None)
    cmd: Optional[str] = field(default=None)
    sw: Optional[str] = field(default=None)
    hw: Optional[str] = field(default=None)
    ack: Optional[str] = field(default=None)
    urls: List[ProtocolOtaUrlJson] = field(default_factory=list)  # this cannot be optional. It throws off dataclass



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

import json
import os.path
import re
from dataclasses import dataclass, field
from os import access, R_OK
from typing import Optional

from .error import DeviceConfigError
from avnet.iotconnect.sdk.sdklib.protocol.files import ProtocolDeviceConfigJson
from ..sdklib.util import deserialize_dataclass


@dataclass
class DeviceConfig:
    """
    This dictionary (dataclass) defines your device's configuration settings.
    You can construct an instance on the fly or use the from_iotc_device_config_json_file class method
    to load most of the device configuration's required parameters from the iotcDeviceConfig.json,
    which you can download by clicking the "note and cog" icon in your device's info panel.
    
    Example:

        device_config = DeviceConfig(
            platform="aws",
            cpid="ABCDEFG",
            env="poc",
            duid="my-device",
            device_cert_path="my-device-cert.pem",
            device_pkey_path="my-device-pkey.pem"
        )

    """
    platform: str = field(default=None)
    """The IoTconnect IoT platform - Either "aws" for AWS IoTCore or "az" for Azure IoTHub"""

    env: str = field(default=None)
    """Your account environment. You can locate this in you IoTConnect web UI at Settings -> Key Value"""

    cpid: str = field(default=None)
    """Your account CPID (Company ID). You can locate this in you IoTConnect web UI at Settings -> Key Value"""

    duid: str = field(default=None)
    """Your device unique ID"""

    device_cert_path: str = field(default=None)
    """Path to the device certificate file"""

    device_pkey_path: str = field(default=None)
    """Path to the device private key file"""

    discovery_url: Optional[str] = field(default=None)
    """Platform - either "aws" or "az" (Azure)"""

    server_ca_cert_path: Optional[str] = field(default=None)  # if not specified use system CA certificates in /etc/ssl or whatever it would be in windows
    """
    Optional path the server certificate that will be used to validate the server connection
    against known CA certificate. If not provided, the system Root CA certificate store (located at /etc/ssl/certs on Linux, for example)
    will be used. If provided, they should be the Amazon Root CA1 AWS, and DigiCert Global Root G2 for Azure.    
    Please note that is more secure to pass the actual server CA Root certificate in order to avoid potential MITM attacks.
    On Linux, you can use server_ca_cert_path="/etc/ssl/certs/DigiCert_Global_Root_CA.pem" for Azure,
    or server_ca_cert_path="/etc/ssl/certs/Amazon_Root_CA_1.pem" for AWS    
    """

    def __post_init__(self):
        """ Validate dataclass arguments and try to infoer some, if they are missing """
        if self.platform not in ("aws", "az"):
            raise DeviceConfigError('DeviceConfig: Platform must be "aws" or "az"')
        if self.discovery_url is None:
            if self.platform == "az":
                self.discovery_url = "https://discovery.iotconnect.io"
            elif self.platform == "aws":
                if self.env == "poc":
                    self.discovery_url = "https://awsdiscovery.iotconnect.io"
                else:
                    # best guess...
                    self.discovery_url = "https://discoveryconsole.iotconnect.io"
        DeviceConfig._validate_file(self.device_cert_path, r"^-----BEGIN CERTIFICATE-----$")
        DeviceConfig._validate_file(self.device_pkey_path, r"^-----BEGIN.*PRIVATE KEY-----$")
        if self.server_ca_cert_path is not None:
            DeviceConfig._validate_file(self.server_ca_cert_path, r"^-----BEGIN CERTIFICATE-----$")

    @classmethod
    def from_iotc_device_config_json(
            cls,
            device_config_json: ProtocolDeviceConfigJson,
            device_cert_path: str,
            device_pkey_path: str,
            server_ca_cert_path: Optional[str] = None) -> 'DeviceConfig':
        """ Return a class instance based on a json string which is in format of the downloadable iotcDeviceConfig.json"""
        if device_config_json.uid is None or device_config_json.cpid is None or device_config_json.env is None or \
                0 == len(device_config_json.uid) or 0 == len(device_config_json.cpid) or 0 == len(device_config_json.env):
            raise DeviceConfigError("The Device Config JSON file format seems to be invalid. Values for cpid, env and uid are required")
        if device_config_json.ver != "2.1":
            raise DeviceConfigError("The Device Config JSON seems to indicate that the device version is not 2.1, which is the only supported version")
        return DeviceConfig(
            env=device_config_json.env,
            cpid=device_config_json.cpid,
            duid=device_config_json.uid,
            platform=device_config_json.pf,
            discovery_url=device_config_json.disc,
            device_cert_path=device_cert_path,
            device_pkey_path=device_pkey_path,
            server_ca_cert_path=server_ca_cert_path
        )

    @classmethod
    def from_iotc_device_config_json_file(
            cls,
            device_config_json_path: str,
            device_cert_path: str,
            device_pkey_path: str,
            server_ca_cert_path: Optional[str] = None) -> 'DeviceConfig':
        """ Return a class instance based on a downloaded iotcDeviceConfig.json fom device's Info panel in IoTConnect"""
        file_content = cls._validate_file(device_config_json_path)
        file_dict = json.loads(file_content)
        pdcj = deserialize_dataclass(ProtocolDeviceConfigJson, file_dict)
        return cls.from_iotc_device_config_json(pdcj, device_cert_path=device_cert_path, device_pkey_path=device_pkey_path, server_ca_cert_path=server_ca_cert_path)

    @classmethod
    def _validate_file(cls, file_name: str, first_line_match_pattern: Optional[str] = None) -> str:
        """ Helper to validate a file - it needs to exist, needs to be readable and (if supplied) the first line has to match the pattern """
        if not os.path.isfile(file_name) or not access(file_name, R_OK):
            raise DeviceConfigError("File %s is not accessible" % file_name)
        file_handle = open(file_name, "r")
        file_content = file_handle.read()
        file_handle.close()
        if file_content is None or 0 == len(file_content):
            raise DeviceConfigError("File %s is empty" % file_name)
        if first_line_match_pattern is not None:
            first_line = file_content.splitlines()[0]
            if not re.search(first_line_match_pattern, first_line):
                raise DeviceConfigError("The file %s does not seem to be valid. Expected the file to start with regex %s" % (file_name, first_line_match_pattern))
        file_handle.close()
        return file_content



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

# The JSON to object mapping was originally created with assistance from OpenAI's ChatGPT.
# For more information about ChatGPT, visit https://openai.com/

from dataclasses import dataclass, field
from typing import Optional


@dataclass
class ProtocolMetaJson:
    at: Optional[int] = None
    df: Optional[int] = None
    cd: Optional[str] = None
    gtw: Optional[int] = None
    edge: Optional[int] = None
    pf: Optional[int] = None
    hwv: str = field(default="")
    swv: str = field(default="")
    v: float = field(default=0.0)


@dataclass
class ProtocolHasJson:
    d: int = field(default=0)
    attr: int = field(default=0)
    set: int = field(default=0)
    r: int = field(default=0)
    ota: int = field(default=0)


@dataclass
class ProtocolSetJson:
    pub: Optional[str] = None
    sub: Optional[str] = None
    pubForAll: Optional[str] = None
    subForAll: Optional[str] = None


@dataclass
class ProtocolTopicsJson:
    rpt: Optional[str] = None
    flt: Optional[str] = None
    od: Optional[str] = None
    hb: Optional[str] = None
    ack: Optional[str] = None
    dl: Optional[str] = None
    di: Optional[str] = None
    c2d: Optional[str] = None
    set: ProtocolSetJson = field(default_factory=ProtocolSetJson)


@dataclass
class ProtocolIdentityPJson:
    n: Optional[str] = None
    h: Optional[str] = None
    p: int = field(default=0)
    id: Optional[str] = None
    un: Optional[str] = None
    topics: ProtocolTopicsJson = field(default_factory=ProtocolTopicsJson)


@dataclass
class ProtocolIdentityDJson:
    ec: int = field(default=0)
    ct: int = field(default=0)
    meta: ProtocolMetaJson = field(default_factory=ProtocolMetaJson)
    has: ProtocolHasJson = field(default_factory=ProtocolHasJson)
    p: ProtocolIdentityPJson = field(default_factory=ProtocolIdentityPJson)
    dt: Optional[str] = None


@dataclass
class ProtocolIdentityResponseJson:
    d: ProtocolIdentityDJson = field(default_factory=ProtocolIdentityDJson)
    status: int = field(default=0)
    message: str = field(default="")



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

# The JSON to object mapping was originally created with assistance from OpenAI's ChatGPT.
# For more information about ChatGPT, visit https://openai.com/


from dataclasses import fields, is_dataclass
from datetime import datetime, timedelta
from typing import get_type_hints, Type, Union, TypeVar


def dict_filter_empty(input_dict: dict):
    return {k: v for k, v in input_dict.items() if v is not None}


def dataclass_factory_filter_empty(data):
    return {key: value for key, value in data if value is not None}


T = TypeVar("T")


def deserialize_dataclass(cls: Type[T], data: Union[dict, list]) -> T:
    """
    Recursively deserialize data into a dataclass or a list of dataclasses.
    """
    if isinstance(data, list):
        # Handle lists of dataclasses
        inner_type = cls.__args__[0] if hasattr(cls, '__args__') else None
        if inner_type and is_dataclass(inner_type):
            return [deserialize_dataclass(inner_type, item) for item in data]
        return data

    if isinstance(data, dict) and is_dataclass(cls):
        field_types = get_type_hints(cls)
        return cls(
            **{
                key: deserialize_dataclass(field_types[key], value)
                if key in field_types and _is_optional_or_dataclass(field_types[key], value)
                else (
                    deserialize_dataclass(field_types[key], value)
                    if key in field_types
                       and hasattr(field_types[key], '__origin__')
                       and field_types[key].__origin__ == list
                    else value
                )
                for key, value in data.items()
                if key in field_types  # Ignore unexpected fields
            }
        )
    return data


def _is_optional_or_dataclass(field_type, value):
    """
    Check if a field type is either an Optional or a dataclass.
    """
    if hasattr(field_type, '__origin__') and field_type.__origin__ is Union:
        # Check for Optional[Type]
        inner_types = field_type.__args__
        if len(inner_types) == 2 and type(None) in inner_types:
            inner_type = [t for t in inner_types if t is not type(None)][0]
            return is_dataclass(inner_type)
    return is_dataclass(field_type)


def filter_init(cls):
    """
    A decorator that modifies the __init__ method of a dataclass to accept a dictionary input.
    It filters out any keys in the input dictionary that are not defined as fields in the dataclass,
    allowing only specified fields to be set during initialization.

    Additionally, if a field is itself a dataclass and is provided as a nested dictionary,
    the decorator will recursively initialize the nested dataclass with the dictionary data.

    Parameters:
    cls : type
        The dataclass to decorate.

    Returns:
    type
        The decorated dataclass with a modified __init__ method.

    Example:
    --------
    @filter_init
    @dataclass
    class Example:
        field1: int
        field2: str

    data = {"field1": 10, "field2": "hello", "extra_field": "ignored"}
    obj = Example(data)  # Initializes Example(field1=10, field2="hello") and ignores "extra_field"
    """

    original_init = cls.__init__

    def __init__(self, input_dict):
        # Get all field names of the dataclass
        field_names = {f.name for f in fields(self)}
        # Filter the input dictionary to keep only the keys that are dataclass fields
        filtered_dict = {k: v for k, v in input_dict.items() if k in field_names}
        # Initialize nested dataclasses if necessary
        for fld in fields(self):
            # Check if the field is a dataclass and its value in input_dict is a dictionary
            if is_dataclass(fld.type) and isinstance(filtered_dict.get(fld.name), dict):
                # Recursively initialize the nested dataclass
                filtered_dict[fld.name] = fld.type(filtered_dict[fld.name])
        # Call the original __init__ with filtered arguments
        original_init(self, **filtered_dict)

    cls.__init__ = __init__
    return cls


class Timing:
    def __init__(self):
        self.t = datetime.now()

    def diff_next(self) -> timedelta:
        now = datetime.now()
        ret = self.diff_with(now)
        self.t = now
        return ret

    def diff_now(self) -> timedelta:
        return datetime.now() - self.t

    def diff_with(self, t: datetime) -> timedelta:
        return t - self.t

    def reset(self, do_print=True) -> timedelta:
        ret = self.diff_next()
        if do_print:
            print("timing: ", ret)
        return ret



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

from dataclasses import dataclass
from datetime import datetime
from typing import Union

# This file contains definitions related to (inbound or outbound) C2D Messages

from avnet.iotconnect.sdk.sdklib.protocol.c2d import ProtocolC2dMessageJson, ProtocolCommandMessageJson, ProtocolOtaUrlJson, ProtocolOtaMessageJson

# When type "object" is defined in IoTConnect, it cannot have nested objects inside of it.
TelemetryValueObjectType = dict[str, Union[None, str, int, float, bool, tuple[float, float]]]
TelemetryValueType = Union[None, str, int, float, bool, tuple[float, float], TelemetryValueObjectType]
TelemetryValues = dict[str, TelemetryValueType]

@dataclass
class TelemetryRecord:
    values: TelemetryValues
    timestamp: datetime = None
    unique_id: str = None
    tag: str = None

class C2dAck:
    """
    OTA Download statuses
    Best practices for OTA status:
    While the final success of the OTA action should be OTA_DOWNLOAD_DONE, it should be generally sent
    only if we are certain that we downloaded the OTA and the new firmware is up and running successfully.
    The user can store the original ACK ID and only report success only after successfully running with the new firmware.
    While the new firmware is downloading over the network or (if available/applicable) writing to the filesystem, unpacking,
    self-testing etc. the intermediate fine-grained statuses along with an appropriate message can be reported
    with the OTA_DOWNLOAD_FAILED status until DONE or FAILED status is reported.
    The exact steps can be left to the interpretation of the user, given the device's capabilities and limitations.
    """

    CMD_FAILED = 1

    CMD_SUCCESS_WITH_ACK  = 2

    # OTA download was not attempted or was rejected.
    OTA_FAILED = 1

    # An intermediate step during the firmware update is pending.
    OTA_DOWNLOADING = 2

    # OTA download is fully completed. New firmware is up and running.
    OTA_DOWNLOAD_DONE = 3

    # The download itself, self-test, writing the files, or unpacking or running the downloaded firmware has failed.
    OTA_DOWNLOAD_FAILED = 4

    @classmethod
    def is_valid_cmd_status(cls, status: int) -> bool:
        return status in (C2dAck.CMD_FAILED, C2dAck.CMD_SUCCESS_WITH_ACK)

    @classmethod
    def is_valid_ota_status(cls, status: int) -> bool:
        return status in (C2dAck.OTA_FAILED, C2dAck.OTA_DOWNLOADING, C2dAck.OTA_DOWNLOAD_DONE, C2dAck.OTA_DOWNLOAD_FAILED)

class C2dMessage:
    COMMAND = 0
    OTA = 1
    MODULE_COMMAND = 2
    REFRESH_ATTRIBUTE = 101
    REFRESH_SETTING = 102
    REFRESH_EDGE_RULE = 103
    REFRESH_CHILD_DEVICE = 104
    DATA_FREQUENCY_CHANGE = 105
    DEVICE_DELETED = 106
    DEVICE_DISABLED = 107
    DEVICE_RELEASED = 108
    STOP_OPERATION = 109
    START_HEARTBEAT = 100
    STOP_HEARTBEAT = 111
    UNKNOWN = 9999

    TYPES: dict[int, str, str] = {
        COMMAND: "Command",
        OTA: "OTA Update",
        REFRESH_ATTRIBUTE: "Refresh Attribute",
        REFRESH_SETTING: "Refresh Setting (Twin/Shadow))",
        REFRESH_EDGE_RULE: "Refresh Edge Rule",
        REFRESH_CHILD_DEVICE: "Refresh Child Device",
        DATA_FREQUENCY_CHANGE: "Data Frequency Changed",
        DEVICE_DELETED: "Device Deleted",
        DEVICE_DISABLED: "Device Disabled",
        DEVICE_RELEASED: "Device Released",
        STOP_OPERATION: "Stop Operation",
        START_HEARTBEAT: "Start Heartbeat",
        STOP_HEARTBEAT: "Stop Heartbeat",
        UNKNOWN: "<Unknown Command Received>"
    }

    def __init__(self, packet: ProtocolC2dMessageJson):
        cls = C2dMessage  # shorthand
        self.type_description = cls.TYPES.get(packet.ct)
        if self.type_description is None:
            self.type_description = cls.TYPES[cls.UNKNOWN]
            self.type = cls.UNKNOWN
        else:
            self.type = packet.ct # this can be None
        self.is_fatal = self.type in (cls.DEVICE_DELETED, cls.DEVICE_DISABLED, cls.DEVICE_RELEASED, cls.STOP_OPERATION)
        self.needs_refresh = self.type in (cls.DATA_FREQUENCY_CHANGE, cls.REFRESH_ATTRIBUTE, cls.REFRESH_SETTING, cls.REFRESH_EDGE_RULE, cls.REFRESH_CHILD_DEVICE)
        self.heartbeat_operation = None
        if self.type == cls.START_HEARTBEAT:
            self.heartbeat_operation = True
        elif self.type == cls.STOP_HEARTBEAT:
            self.heartbeat_operation = False
        self.frequency = packet.f or packet.df  # pick up frequency with respect to DATA_FREQUENCY_CHANGE or Heartbeat

    def validate(self) -> bool:
        return self.type is not None


class C2dCommand:
    def __init__(self, packet: ProtocolCommandMessageJson):
        self.type = C2dMessage.COMMAND  # Used for error checking when sending ACKs
        self.ack_id = packet.ack
        if packet.cmd is not None and len(packet.cmd) > 0:
            cmd_split = packet.cmd.split()
            self.command_name = cmd_split[0]
            self.command_args = cmd_split[1:]
            self.command_raw = packet.cmd
        else:
            self.command_name = None
            self.command_args = []
            self.command_raw = ""

    def validate(self) -> bool:
        return self.command_name is not None


class C2dOta:
    class Url:
        def __init__(self, entry: ProtocolOtaUrlJson):
            self.url = entry.url
            self.file_name = entry.fileName

    def __init__(self, packet: ProtocolOtaMessageJson):
        self.type = C2dMessage.OTA  # Used for error checking when sending ACKs
        self.ack_id = packet.ack
        self.version = packet.sw  # along with OTA superset
        self.hardware_version = packet.hw
        if packet.urls is not None and len(packet.urls) > 0:
            self.urls: list[C2dOta.Url] = [C2dOta.Url(x) for x in packet.urls]
        else:
            # we will let the client handle this case
            self.urls: list[C2dOta.Url] = []

    def validate(self) -> bool:
        if len(self.urls) == 0: return False
        if self.ack_id is None or len(self.ack_id) == 0: return False # OTA must have ack ID
        for u in self.urls:
            if u is None: return False
            if u.file_name is None or len(u.file_name) == 0: return False
            if u.url is None or len(u.url) == 0: return False
        return True



__version__ = '1.0.3'

from .error import DeviceConfigError
from .config import DeviceConfig
from .client import Client, ClientSettings, Callbacks

# redirect these imports so that the user code is not affected by any changes in file organization
from avnet.iotconnect.sdk.sdklib.mqtt import C2dCommand, C2dOta, C2dAck, C2dMessage, TelemetryRecord



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

import urllib.request
from urllib.error import HTTPError, URLError

from avnet.iotconnect.sdk.sdklib.dra import DraDiscoveryUrl, DraDeviceInfoParser, DraIdentityUrl, \
    DeviceIdentityData
from .config import DeviceConfig
from .error import DeviceConfigError


class DeviceRestApi:
    def __init__(self, config: DeviceConfig):
        self.config = config

    def get_identity_data(self) -> DeviceIdentityData:
        try:
            print("Requesting Discovery Data %s..." % DraDiscoveryUrl(self.config).get_api_url())
            resp = urllib.request.urlopen(urllib.request.Request(DraDiscoveryUrl(self.config).get_api_url()))
            discovery_base_url = DraDeviceInfoParser.parse_discovery_response(resp.read())

            print("Requesting Identity Data %s..." % DraIdentityUrl(discovery_base_url).get_uid_api_url(self.config))
            resp = urllib.request.urlopen(DraIdentityUrl(discovery_base_url).get_uid_api_url(self.config))
            identity_response = DraDeviceInfoParser.parse_identity_response(resp.read())
            return identity_response

        except HTTPError as http_error:
            raise DeviceConfigError(http_error)

        except URLError as url_error:
            raise DeviceConfigError(str(url_error))



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

# The JSON to object mapping was originally created with assistance from OpenAI's ChatGPT.
# For more information about ChatGPT, visit https://openai.com/

# This file contains outbound message types (while some messages are related to inbound C2D Messages)

from dataclasses import dataclass, field
from typing import Optional, Any


@dataclass
class ProtocolTelemetryEntryJson:
    d: dict[str, Any] = field(default_factory=dict)
    dt: Optional[str] = field(default=None)
    id: Optional[str] = field(default=None)
    tg: Optional[str] = field(default=None)


@dataclass
class ProtocolTelemetryMessageJson:
    d: list[ProtocolTelemetryEntryJson] = field(default_factory=list[ProtocolTelemetryEntryJson])
    # there is also a top level timestamp which has dubious meaning, so we don't support it

@dataclass
class ProtocolAckDJson:
    ack: Optional[str] = field(default=None)
    type: Optional[int] = field(default=None)
    st: Optional[int] = field(default=None)
    msg: Optional[str] = field(default=None)

@dataclass
class ProtocolAckMessageJson:
    d: ProtocolAckDJson = field(default_factory=ProtocolAckDJson)
    dt: Optional[str] = field(default=None)



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

# The JSON to object mapping was originally created with assistance from OpenAI's ChatGPT.
# For more information about ChatGPT, visit https://openai.com/

from dataclasses import dataclass, field
from typing import Optional


@dataclass
class ProtocolDeviceConfigJson:
    cpid: Optional[str] = field(default=None)
    env: Optional[str] = field(default=None)
    uid: Optional[str] = field(default=None)
    did: Optional[str] = field(default=None)
    disc: Optional[str] = field(default=None)
    ver: Optional[str] = field(default="2.1")
    pf: Optional[str] = field(default="aws")
    at: Optional[int] = field(default=7)  # authentication type, but ignored for now



# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

class DeviceConfigError(Exception):
    def __init__(self, message: str):
        self.msg = message
        super().__init__(message)

class ClientError(Exception):
    def __init__(self, message: str):
        self.msg = message
        super().__init__(message)


# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Avnet
# Authors: Nikola Markovic <nikola.markovic@avnet.com> et al.

import json
import random
import time
from dataclasses import asdict
from datetime import datetime, timezone
from json import JSONDecodeError
from ssl import SSLError
from typing import Callable, Optional

from paho.mqtt.client import CallbackAPIVersion, MQTTErrorCode, DisconnectFlags, MQTTMessageInfo
from paho.mqtt.client import Client as PahoClient
from paho.mqtt.reasoncodes import ReasonCode

from avnet.iotconnect.sdk.sdklib.mqtt import C2dOta, C2dMessage, C2dCommand, C2dAck, TelemetryRecord, TelemetryValueType
from avnet.iotconnect.sdk.sdklib.protocol.c2d import ProtocolC2dMessageJson, ProtocolOtaMessageJson, ProtocolCommandMessageJson
from avnet.iotconnect.sdk.sdklib.protocol.d2c import ProtocolTelemetryMessageJson, ProtocolTelemetryEntryJson, ProtocolAckDJson, ProtocolAckMessageJson
from avnet.iotconnect.sdk.sdklib.util import Timing, dataclass_factory_filter_empty, deserialize_dataclass
from .config import DeviceConfig
from .dra import DeviceRestApi


class Callbacks:
    """
    Specify callbacks for C2D command, OTA (not implemented yet) or MQTT disconnection.

    :param command_cb: Callback function with first parameter being C2dCommand object.
        Use this callback to process commands sent by the back end.

    :param ota_cb: Callback function with first parameter being C2dOta object.
        Use this callback to process OTA updates sent by the back end.

    :param disconnected_cb: Callback function with first parameter being string with reason for disconnect
        and second a boolean indicating whether a disconnect request was received from the server.
        Use this callback to asynchronously react to the back end disconnection event rather than polling Client.is_connected.
    """

    def __init__(
            self,
            command_cb: Optional[Callable[[C2dCommand], None]] = None,
            ota_cb: Optional[Callable[[C2dOta], None]] = None,
            disconnected_cb: Optional[Callable[[str, bool], None]] = None,
            generic_message_callbacks: dict [int, Callable[[C2dMessage, dict], None]]= None
    ):
        self.disconnected_cb = disconnected_cb
        self.command_cb = command_cb
        self.ota_cb = ota_cb
        self.generic_message_callbacks = generic_message_callbacks or dict[int, C2dMessage]() # empty dict otherwise


class ClientSettings:
    """ Optional settings that the user can use to control the client MQTT connection behavior"""

    def __init__(
            self,
            verbose: bool = True,
            connect_timeout_secs: int = 30,
            connect_tries: int = 100,
            connect_backoff_max_secs: int = 15
    ):
        self.verbose = verbose
        self.connect_timeout_secs = connect_timeout_secs
        self.connect_tries = connect_tries
        self.connect_backoff_max_secs = connect_backoff_max_secs
        if connect_timeout_secs < 1:
            raise ValueError("connect_timeout_secs must be greater than 1")
        if connect_tries < 1:
            raise ValueError("connect_tries must be greater than 1")


class Client:
    """
    This is an Avnet IoTConnect MQTT client that provides an easy way for the user to
    connect to IoTConnect, send and receive data.

    :param config: Required device configuration. See the examples or DeviceConfig class description for more details.
    :param callbacks: Optional callbacks that can be provided for the following events:
        - IoTConnect C2D (Cloud To Device) commands that you can send to your device using the IoTConnect
        - IoTConnect OTA update events.
        - Device MQTT disconnection.
    :param settings: Tune the client behavior by providing your preferences regarding connection timeouts and logging.

    Usage (see basic-example.py or minimal.py examples at https://github.com/avnet-iotconnect/iotc-python-lite-sdk for more details):

    - Construct this client class:
        - Provide your device information and x509 credentials (certificate and private key)
            Either provide the path the downloaded iotcDeviceConfig.json and use the DeviceConfig.from_iotc_device_config_json_file()
            class method. You can download the iotcDeviceConfig.json by clicking the cog icon in the upper right of your device's info panel.

            Or you can also provide the device parameters directly:

            device_config = DeviceConfig(
                platform="aws",                             # The IoTconnect IoT platform - Either "aws" for AWS IoTCore or "az" for Azure IoTHub
                env="your-environment",                     # Your account environment. You can locate this in you IoTConnect web UI at Settings -> Key Value
                cpid="ABCDEFG123456",                       # Your account CPID (Company ID). You can locate this in you IoTConnect web UI at Settings -> Key Value
                duid="my-device",                           # Your device unique ID
                device_cert_path="path/to/device-cert.pem", # Path to the device certificate file
                device_pkey_path="path/to/device-pkey.pem"  # Path to the device private key file
            )
            NOTE: If you do not pass the server certificate, we will use the system's trusted certificate store, if available.
            For example, the trusted Root CA certificates from the in /etc/ssl/certs will be used on Linux.
            However, it is more secure to pass the actual server CA Root certificate in order to avoid potential MITM attacks.
            On Linux, you can use server_ca_cert_path="/etc/ssl/certs/DigiCert_Global_Root_CA.pem" for Azure
            or server_ca_cert_path="/etc/ssl/certs/Amazon_Root_CA_1.pem" for AWS

        - (Optional) provide callbacks for C2D Commands or device disconnect.
            See the basic-example.py at https://github.com/avnet-iotconnect/iotc-python-lite-sdk example for details.

        - (Optional) provide ClientSettings to tune the client behavior.

        It is recommended to surround the DeviceConfig constructor and the Client constructor
        with a try/except block catching DeviceConfigError and printing it to get more information
        to be able to troubleshoot configuration errors.


    - Call Client.connect(). The call will block until successfully connected or timed out. For example:

    - Ensure that the Client.is_connected() periodically or react to a disconnect event callback
        and attempt to reconnect or perform a different action appropriate to your application. For example:
        if not client.is_connected:
            client.connect()

    - Send messages with Client.send_telemetry() or send_telemetry_records().
        See basic-example.py example at https://github.com/avnet-iotconnect/iotc-python-lite-sdk for more info.
        For example:
            c.send_telemetry({
                'temperature': get_sensor_temperature()
            })

    """

    def __init__(
            self,
            config: DeviceConfig,
            callbacks: Callbacks = None,
            settings: ClientSettings = None
    ):
        self.config = config
        self.user_callbacks = callbacks or Callbacks()
        self.settings = settings or ClientSettings()

        self.mqtt_config = DeviceRestApi(config).get_identity_data()  # can raise DeviceConfigError

        self.mqtt = PahoClient(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=self.mqtt_config.client_id
        )
        # TODO: User configurable with defaults
        self.mqtt.reconnect_delay_set(min_delay=1, max_delay=int(self.settings.connect_timeout_secs / 2 + 1))
        self.mqtt.tls_set(certfile=config.device_cert_path, keyfile=config.device_pkey_path, ca_certs=config.server_ca_cert_path)
        self.mqtt.username = self.mqtt_config.username

        self.mqtt.on_message = self._on_mqtt_message
        self.mqtt.on_connect = self._on_mqtt_connect
        self.mqtt.on_disconnect = self._on_mqtt_disconnect
        self.mqtt.on_publish = self._on_mqtt_publish

        self.user_callbacks = callbacks or Callbacks()

    @classmethod
    def timestamp_now(cls) -> datetime:
        """ Returns the UTC timestamp that can be used to stamp telemetry records """
        return datetime.now(timezone.utc)

    def is_connected(self):
        return self.mqtt.is_connected()

    def connect(self):
        def wait_for_connection() -> bool:
            connect_timer = Timing()
            if self.settings.verbose:
                print("waiting to connect...")
            while True:
                if self.is_connected():
                    if self.settings.verbose:
                        print("MQTT connected")
                    return True
                time.sleep(0.5)
                if connect_timer.diff_now().seconds > self.settings.connect_timeout_secs:
                    print("Timed out.")
                    self.disconnect()
                    return False

        if self.is_connected():
            return

        for i in range(1, self.settings.connect_tries):
            try:
                t = Timing()
                mqtt_error = self.mqtt.connect(
                    host=self.mqtt_config.host,
                    port=8883
                )
                if mqtt_error != MQTTErrorCode.MQTT_ERR_SUCCESS:
                    print("TLS connection to the endpoint failed")
                else:
                    print("Awaiting MQTT connection establishment...")
                    self.mqtt.loop_start()
                    if wait_for_connection():
                        if self.settings.verbose:
                            print("Connected in %dms" % (t.diff_now().microseconds / 1000))
                        break
                    else:
                        continue

            except (SSLError, TimeoutError, OSError) as ex:
                # OSError includes socket.gaierror when host could not be resolved
                # This could also be temporary, so keep trying
                print("Failed to connect to host %s. Exception: %s" % (self.mqtt_config.host, str(ex)))

            backoff_ms = random.randrange(1000, self.settings.connect_backoff_max_secs * 1000)
            print("Retrying connection... Backing off for %d ms." % backoff_ms)
            # Jitter back off a random number of milliseconds between 1 and 10 seconds.
            time.sleep(backoff_ms / 1000)

        self.mqtt.subscribe(self.mqtt_config.topics.c2d, qos=1)

    def disconnect(self) -> MQTTErrorCode:
        ret = self.mqtt.disconnect()
        if self.settings.verbose:
            print("Disconnected.")
        return ret

    def send_telemetry(self, values: dict[str, TelemetryValueType], timestamp: datetime = None):
        """ Sends a single telemetry dataset. 
        If you need gateway/child functionality or need to send multiple value sets in one packet, 
        use the send_telemetry_records() method.
         
        :param TelemetryValues values:
            The name-value telemetry pairs to send. Each value can be
                - a primitive value: Maps directly to a JSON string, number or boolean
                - None: Maps to JSON null,
                - Tuple[float, float]: Used to send a lat/long geographic coordinate as decimal degrees as an
                    array of two (positive or negative) floats.
                    For example, [44.787197, 20.457273] is the geo coordinate Belgrade in Serbia,
                    where latitude 44.787197 is a positive number indicating degrees north,
                    and longitude a positive number as well, indicating degrees east.
                    Maps to JSON array of two elements.
                - Another hash with possible values above when sending an object. Maps to JSON object.
            in case when an object needs to be sent.
        :param datetime timestamp: (Optional) The timestamp corresponding to this dataset.
            If not provided, this will save bandwidth, as no timestamp will not be sent over MQTT.
             The server receipt timestamp will be applied to the telemetry values in this telemetry record.
             Supply this value (using Client.timestamp()) if you need more control over timestamps.
        """
        self.send_telemetry_records([TelemetryRecord(
            values=values,
            timestamp=timestamp
        )])

    def send_telemetry_records(self, records: list[TelemetryRecord]) -> Optional[MQTTMessageInfo]:
        """
        A complex, but more powerful way to send telemetry.
        It allows the user to send multiple sets of telemetry values
        and control the timestamp of each telemetry value set.
        Supports gateway devices with gateway-child relationship by allowing
        the user to set the parent/child unique_id ("id" in JSON)
        and tag of respective parent./child ("tg" in JSON)

        See https://docs.iotconnect.io/iotconnect/sdk/message-protocol/device-message-2-1/d2c-messages/#Device for more information.
        """

        if not self.is_connected():
            print('Message NOT sent. Not connected!')
            return None
        else:
            packet = ProtocolTelemetryMessageJson()
            for r in records:
                packet_entry = ProtocolTelemetryEntryJson(
                    d=r.values,
                    dt=None if r.timestamp is None else Client._to_iotconnect_time_str(r.timestamp),
                    id=r.unique_id,
                    tg=r.tag
                )
                packet.d.append(asdict(packet_entry, dict_factory=dataclass_factory_filter_empty))
            iotc_json = json.dumps(asdict(packet), separators=(',', ':'))

            ret = self.mqtt.publish(
                topic=self.mqtt_config.topics.rpt,
                qos=1,
                payload=iotc_json
            )
            if self.settings.verbose:
                print(">", iotc_json)
            return ret


    def send_command_ack(self, original_message: C2dCommand, status: int, message_str = None):
        """
        Send Command acknowledgement.

        :param original_message: The original message that was received in the callback.
        :param status: C2dAck.CMD_FAILED or C2dAck.CMD_SUCCESS_WITH_ACK.
        :param message_str: (Optional) For example: 'LED color now "blue"', or 'LED color "red" not supported'
        """

        if original_message.type != C2dMessage.COMMAND:
            print('Error: Called send_command_ack(), but message is not a command!')
            return
        self.send_ack(
            ack_id=original_message.ack_id,
            message_type=original_message.type,
            status=status,
            message_str=message_str,
            original_command=original_message.command_name
        )

    def send_ota_ack(self, original_message: C2dOta, status: int, message_str = None):
        """
        Send OTA acknowledgement.
        See the C2dAck comments for best practices with OTA download ACks.

        :param original_message: The original message that was received in the callback.
        :param status: For example: C2dAck.OTA_DOWNLOAD_FAILED.
        :param message_str: (Optional) For example: "Failed to unzip the OTA package".
        :param message_str: (Optional) For example: "Failed to unzip the OTA package".
        """
        if original_message.type != C2dMessage.OTA:
            print('Error: Called send_ota_ack(), but message is not an OTA request!')
            return
        self.send_ack(
            ack_id=original_message.ack_id,
            message_type=original_message.type,
            status=status,
            message_str=message_str
        )

    def send_ack(self, ack_id: str, message_type: int, status: int, message_str: str = None, original_command: str = None):
        """
        Send Command or OTA ack while having only ACK ID

        :param ack_id: Recorded ack_id from the original message.
        :param message_type: For example: C2dMessage.COMMAND or C2dMessage.OTA, .
        :param status: For example: C2dAck.CMD_FAILED or C2dAck.OTA_DOWNLOAD_DONE for command/OTA respectively.
        :param message_str: (Optional) For example: "Failed to unzip the OTA package", or 'LED color "red" not supported'
        :param original_command: (Optional) If this argument is passed,
            this command name will be printed along with any potential error messages.

        While the client should generally use send_ota_ack or send_command_ack, this method can be used in cases
        where the context of the original received message is not available (after OTA restart for example)
        """
        if not self.is_connected():
            print('Message NOT sent. Not connected!')
        elif ack_id is None or len(ack_id) == 0:
            if original_command is not None:
                print('Error: Message ACK ID missing. Ensure to set "Acknowledgement Required" in the template for command %s!' % original_command)
            else:
                print('Error: Message ACK ID missing. Ensure to set "Acknowledgement Required" in the template the command!' % original_command)
            return
        elif message_type not in (C2dMessage.COMMAND, C2dMessage.OTA):
            print('Warning: Message type %d does not appear to be a valid message type!' % message_type)  # let it pass, just in case we can still somehow send different kind of ack
        elif message_type == C2dMessage.COMMAND and not C2dAck.is_valid_cmd_status(status):
            print('Warning: Status %d does not appear to be a valid command ACK status!' % status) # let it pass, just in case there is a new status
        elif message_type == C2dMessage.OTA and not C2dAck.is_valid_ota_status(status):
            print('Warning: Status %d does not appear to be a valid OTA ACK status!' % status) # let it pass, just in case there is a new status

        packet = ProtocolAckMessageJson(
            d=ProtocolAckDJson(
                ack=ack_id,
                type=message_type,
                st=status,
                msg=message_str
            )
        )
        iotc_json = json.dumps(asdict(packet, dict_factory=dataclass_factory_filter_empty), separators=(',', ':'))
        ret = self.mqtt.publish(
            topic=self.mqtt_config.topics.ack,
            qos=1,
            payload=iotc_json
        )
        if self.settings.verbose:
            print(">", iotc_json)
        return ret

    @classmethod
    def _to_iotconnect_time_str(cls, ts: datetime) -> str:
        return ts.strftime("%Y-%m-%dT%H:%M:%S.000Z")

    def _process_c2d_message(self, topic: str, payload: str) -> bool:
        # topic is ignored for now as we only subscribe to one
        # we ought to change this once we start supporting Properties (Twin/Shadow)
        try:
            # use the simplest form of ProtocolC2dMessageJson when deserializing first and
            # convert message to appropriate json later
            message_dict = json.loads(payload)
            message_packet = deserialize_dataclass(ProtocolC2dMessageJson, message_dict)
            message = C2dMessage(message_packet)

            if not message.validate():
                print("C2D Message is invalid: %s" % payload)
                return False

            # if the user wants to handle this message type, stop processing further
            generic_cb = self.user_callbacks.generic_message_callbacks.get(message.type)
            if generic_cb is not None:
                generic_cb(message, message_dict)
                return True

            if message.type == C2dMessage.COMMAND:
                msg = C2dCommand(deserialize_dataclass(ProtocolCommandMessageJson, message_dict))
                if not msg.validate():
                    print("C2D Command is invalid: %s" % payload)
                    return False
#               TODO: Deal with runtime qualification
#               if msg.command_name == 'aws-qualification-start':
#                    self._aws_qualification_start(msg.command_args)
#                elif self.user_callbacks.command_cb is not None:
                if self.user_callbacks.command_cb is not None:
                    self.user_callbacks.command_cb(msg)
                else:
                    if self.settings.verbose:
                        print("WARN: Unhandled command %s received!" % msg.command_name)
            elif message.type == C2dMessage.OTA:
                msg = C2dOta(deserialize_dataclass(ProtocolOtaMessageJson, message_dict))
                if not msg.validate():
                    print("C2D Command is invalid: %s" % payload)
                    return False
                if len(msg.urls) > 0:
                    if self.user_callbacks.ota_cb is not None:
                        self.user_callbacks.ota_cb(msg)
                    else:
                        if self.settings.verbose:
                            print("WARN: Unhandled OTA request received!")
                else:
                    print("ERROR: Got OTA, but URLs list is empty!")
            elif message.is_fatal:
                print("Received C2D message %s from backend. Device should stop operation." % message.type_description)
            elif message.needs_refresh:
                print("Received C2D message %s from backend. Device should re-initialize the application." % message.type_description)
            elif message.heartbeat_operation is not None:
                operation_str = "start" if message.heartbeat_operation == True else "stop"
                print("Received C2D message %s from backend. Device should %s heartbeat messages." % (message.type_description, operation_str))
            else:
                print("C2D Message parsing for message type %d is not supported by this client. Message was: %s" % (message_packet.ct, payload))

        except JSONDecodeError:
            print('Error: Incoming message not parseable: "%s"' % payload)
            return False

    def _on_mqtt_connect(self, mqttc: PahoClient, obj, flags, reason_code, properties):
        if self.settings.verbose:
            print("Connected. Reason Code: " + str(reason_code))

    def _on_mqtt_disconnect(self, mqttc: PahoClient, obj, flags: DisconnectFlags, reason_code: ReasonCode, properties):
        if self.user_callbacks.disconnected_cb is not None:
            # cannot send raw reason code from paho. We could technically change the backend.
            self.user_callbacks.disconnected_cb(str(reason_code), flags.is_disconnect_packet_from_server)
        else:
            print("Disconnected. Reason: %s. Flags: %s" % (str(reason_code), str(flags)))

    def _on_mqtt_message(self, mqttc: PahoClient, obj, msg):
        if self.settings.verbose:
            print(msg.topic + " " + str(msg.qos) + " " + str(msg.payload))
        self._process_c2d_message(msg.topic, msg.payload)

    def _on_mqtt_publish(self, mqttc: PahoClient, obj, mid, reason_code, properties):
        # print("mid: " + str(mid))
        pass

    def _aws_qualification_start(self, command_args: list[str]):
        t = Timing()

        def log_callback(client, userdata, level, buf):
            print("%d [%s]: %s" % (t.diff_now().microseconds / 1000, str(level), buf))
            t.reset(False)

        if len(command_args) >= 1:
            host = command_args[0]
            print("Starting AWS Device Qualification for", host)
            self.mqtt_config.topics.rpt = 'qualification'
            self.mqtt_config.topics.c2d = 'qualification'
            self.mqtt_config.topics.ack = 'qualification'
            self.mqtt_config.host = host
            self.mqtt.on_log = log_callback
            self.disconnect()
            while True:
                connected_time = Timing()
                if not self.is_connected():
                    print('(re)connecting to', self.mqtt_config.host)
                    self.connect()
                    connected_time.reset(False)  # reset the timer
                else:
                    if connected_time.diff_now().seconds > 60:
                        print("Stayed connected for too long. resetting the connection")
                        self.disconnect()
                        continue
                    self.send_telemetry({
                        'qualification': 'true'
                    })
                time.sleep(5)

        else:
            print("Malformed AWS qualification command. Missing command argument!")
